/*
 * Volleyball - video game similary to GNU Arcade Volleyball
 * Copyright (C) 2005, 2006 Hugo Ruscitti : hugoruscitti@gmail.com
 * web site: http://www.losersjuegos.com.ar
 * 
 * This file is part of Volleyball.
 *
 * Volleyball is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Volleyball is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

#include "ball.h"

/*!
 * \brief inicia el mdulo gestor de la pelota
 * \param position indica cual es el jugador que tiene el control inicial de
 * la pelota, los valores vlidos aqu sern 0 o 1
 * \param p1 jugador 1
 * \param p2 jugador 2
 * \param status_bar barra con el puntaje y tiempo de juego
 * \param debug indica si se deben imprimir las lineas gua
 */
Ball * ball_init (int position, Player * p1, Player * p2, 
		Status_bar * status_bar, int debug)
{
	Ball * data;

	data = (Ball *) malloc (sizeof (Ball));

	if (! data)
	{
		util_print_fail_module ("Ball");
		exit (1);
	}
	
	data->step = 0;
	data->collision = 0;
	
	data->image = util_load_image (DATADIR "/ball.png", FALSE);

	if (! data->image)
	{
		util_print_fail_module ("Ball");
		exit (1);
	}

	data->collision_image = util_load_image (DATADIR "/ball_collision.png",
			FALSE);

	if (! data->collision_image)
	{
		util_print_fail_module ("Ball");
		exit (1);
	}

	data->player_1 = p1;
	data->player_2 = p2;
	data->status_bar = status_bar;
	data->ratio = data->image->h / 2;
	data->vx = 0.0;
	data->vy = 0.0;
	data->power = 3.0;
	data->player_start = util_rand (0, 1);
	
	ball_change_state (data, ball_init_restart, ball_state_restart);

	if (debug)
		ball_draw_debug_lines (data);

	return data;
}


/*!
 * \brief Actualizacin lgica
 * \param data mdulo principal
 */
void ball_update (Ball * data)
{
	data->state (data);
	ball_animation_update (data);
	
	if (data->collision)
		data->collision --;
}


/*!
 * \brief Actualiza la animacin de la pelota
 * \param data mdulo principal
 */
void ball_animation_update (Ball * data)
{
	data->step += data->vx / 4;

	if (data->step < 0.0)
		data->step = 23.0;
	else
	{
		if (data->step > 23.0)
			data->step = 0;
	}
}


/*!
 * \brief imprime la pelota en pantalla
 * \param data mdulo principal
 * \param display gestor de pantalla
 */
void ball_draw (Ball * data, Display * display)
{
	int x = data->x - data->ratio;
	int y = data->y - data->ratio;
	SDL_Surface * image;

	if (data->collision)
		image = data->collision_image;
	else
		image = data->image;

	display_draw (display, image, data->step, 24, x, y, FALSE);
}


/*!
 * \brief termina con el mdulo
 * \param data mdulo principal
 */
void ball_free (Ball * data)
{
	SDL_FreeSurface (data->image);
	SDL_FreeSurface (data->collision_image);
	free (data);
}

/*!
 * \brief inicia el estado `bouncing` (rebotando)
 * \param data mdulo principal
 */
void ball_init_bouncing (Ball * data)
{
}


/*!
 * \brief inicia el estado `restart` (reinicia)
 * \param data mdulo principal
 */
void ball_init_restart (Ball * data)
{
	data->delay = 128;
	data->vx = 0;

	data->final_y = 280;
	
	if (data->player_start)
		data->final_x = 320 + 160;
	else
		data->final_x = 160;

	data->angle = 0.0;
}


/*!
 * \brief inicia el estado `ready` (esperando a iniciar)
 * \param data mdulo principal
 */
void ball_init_ready (Ball * data)
{
	data->power = 3;
}

/*!
 * \brief inicia el estado `goal` (cuando la pelota toca el suelo)
 * \param data mdulo principal
 */
void ball_init_goal (Ball * data)
{
	int player_win;
	
	if (data->x < 320)
		player_win = 1;
	else
		player_win = 0;

	/* solo aumenta un tanto al jugador si fu el que realiz el saque */
	if (data->player_start == player_win)
		status_bar_goal (data->status_bar, player_win);

	data->player_start = player_win;
	
	data->delay = 128;
}


/*!
 * \brief gestiona el estado `goal` (mientras la pelota desaparece)
 * \param data mdulo principal
 */
void ball_state_goal (Ball * data)
{
	int flags = SDL_RLEACCEL | SDL_SRCALPHA;

	data->x += data->vx;
	data->y += data->vy;
	
	data->vy += 0.04;
	
	ball_check_boundary_collision (data);
	
	if (data->delay > 0)
	{
		data->delay --;
		SDL_SetAlpha (data->image, flags, data->delay * 2);
	}
	else
	{
		ball_change_state (data, ball_init_restart, ball_state_restart);
	}
}

/*!
 * \brief gestor del estado `bouncing` (rebotando)
 * \param data mdulo principal
 */
void ball_state_bouncing (Ball * data)
{
	data->x += data->vx;
	data->y += data->vy;

	data->vy += 0.04;
	
	if (data->vy > 0.0 && data->collision < 1)
		ball_collision (data);

	/* si la pelota toca el suelo ... */
	if (ball_check_boundary_collision (data))
	{
		ball_change_state (data, ball_init_goal, ball_state_goal);
	}
}


/*!
 * \brief gestor del estado `restart` (reinicia)
 * \param data mdulo principal
 */
void ball_state_restart (Ball * data)
{
	int flags = SDL_RLEACCEL | SDL_SRCALPHA;

	if (data->delay < 1)
	{
		ball_change_state (data, ball_init_ready, ball_state_ready);
	}
	else
	{	
		data->delay --;
		data->angle += 0.1;

		data->x = cos (data->angle) * data->delay + data->final_x;
		data->y = sin (data->angle) * data->delay + data->final_y;

		SDL_SetAlpha (data->image, flags, 255 - data->delay * 2);
	}
}


/*!
 * \brief gestor del estado `ready` (esperando iniciar)
 * \param data mdulo principal
 */
void ball_state_ready (Ball * data)
{
	if (ball_collision (data))
	{
		ball_change_state (data, ball_init_bouncing, 
				ball_state_bouncing);
	}
}


/*!
 * \brief cambia el estado actual de la pelota
 * \param mdulo principal
 * \param init funcin de inicio de estado
 * \param update funcin que gestiona la actualizacin del estado
 */
void ball_change_state (Ball * data, void (* init) (Ball * data), 
		void (* update) (Ball * data))
{
	if (init)
		init (data);

	data->state = update;
}

/*!
 * \brief Imprime lineas guias para el modo depuracin
 * \param data mdulo principal
 */
void ball_draw_debug_lines (Ball * data)
{
	Uint32 color = SDL_MapRGB (data->image->format, 255, 120, 120);
	int i;
	int dx = 0;
	int ratio = data->ratio;

	for (i = 0; i < 24; i ++)
	{
		util_draw_circle (data->image, color, dx + ratio, ratio, ratio);
		dx += ratio * 2;
	}
}


/*!
 * \brief Verifica si la pelota se encuentra en colisin con los personajes
 * \param data mdulo principal
 * \return TRUE si colision con un personaje
 */
int ball_collision (Ball * data)
{
	int ratio = data->ratio;
	int px;
	int py;
	int pratio;
	Player * player;

	if (data->x < 320)
		player = data->player_1;
	else
		player = data->player_2;

	player_get_collision_area (player, & px, & py, & pratio);

	if (util_is_collision (data->x, data->y, px, py, pratio + ratio))
	{
		float angle;
		int dx = data->x - px;

		angle = util_get_angle (data->x, data->y, px, py);
	
		data->power *= player_get_collision_power (player, dx);
	
		/* limita la potencia de la pelota */
		if (data->power > 7.0)
			data->power = 7.0;
		else
		{
			if (data->power < 2.0)
				data->power = 2.0;
		}
		
		data->vx = cos (angle) * data->power;
		data->vy = - sin (angle) * data->power;

		ball_separate (data, px, py, pratio);
			
		data->collision = 25;

		/* informa al jugador que ha colisionado con la pelota */
		player_collision_with_ball (player, dx);
		
		return TRUE;
	}

	return FALSE;
}


/*!
 * \brief evita que la pelota salga de la pantalla
 * \param data mdulo principal
 * \return TRUE si la pelota toca el suelo
 */
int ball_check_boundary_collision (Ball * data)
{
	/* verifica los lmites derecho e izquierdo */
	if (data->vx > 0)
	{
		if (data->x + data->vx + data->ratio > 640)
		{
			data->x = 640 - data->ratio;
			data->vx *= -1;
			return FALSE;
		}
	}
	else
	{
		if (data->x + data->vx - data->ratio < 0)
		{
			data->x = data->ratio;
			data->vx *= -1;
			return FALSE;
		}
	}

	/* verifica los lmites superior e inferior */
	if (data->vy < 0)
	{
		if (data->y + data->vy - data->ratio < 0)
		{
			data->y = data->ratio;
			data->vy *= -1;
			return FALSE;
		}
	}
	else
	{
		if (data->y + data->vy + data->ratio > FLOOR)
		{
			data->y = FLOOR - data->ratio;
			data->vy /= -2;
			data->vx /= 2;
			return TRUE;
		}
	}

	/* verifica lmites con la barra central */
	if (data->y < PILLAR_Y)
	{
		if (util_is_collision (data->x, data->y, PILLAR_X, PILLAR_Y, 
					data->ratio + PILLAR_RATIO))
		{
			float angle;
			
			angle = util_get_angle (data->x, data->y, PILLAR_X, 
					PILLAR_Y);
			
			ball_separate (data, PILLAR_X, PILLAR_Y, PILLAR_RATIO);

			data->vx = cos (angle) * data->power * 0.8;
			data->vy = - sin (angle) * data->power * 0.8;

			if (data->vx < 0.25 && data->vx > -0.25)
			{
				if (data->vx < 0.0)
					data->vx = -0.25;
				else
					data->vx = 0.25;
			}

		}
	}
	else
	{
		/* colisiona con el poste del lado izquierdo */
		if (util_include_value (data->x + data->ratio, 
					PILLAR_X - PILLAR_RATIO,
					PILLAR_X + PILLAR_RATIO))
		{
			data->x = PILLAR_X - PILLAR_RATIO - data->ratio - 1;
			data->vx *= -1;
			return FALSE;
		}
		
		/* colisiona con el poste del lado derecho */
		if (util_include_value (data->x - data->ratio, 
					PILLAR_X - PILLAR_RATIO,
					PILLAR_X + PILLAR_RATIO))
		{
			data->x = PILLAR_X + PILLAR_RATIO + data->ratio + 1;
			data->vx *= -1;
			return FALSE;
		}
						
	}

	
	return FALSE;
}

/*!
 * \brief Evita que la pelota y una circunferencia se solapen
 * \param data mdulo principal
 * \param px posicin coordenada horizontal de la circunferencia
 * \param py posicin coordenada vertical de la circunferencia
 * \param pratio radio de la circunferencia
 */
void ball_separate (Ball * data, int px, int py, int pratio)
{
	float angle = util_get_angle (data->x, data->y, px, py);
	int d;

	/* separa la pelota del personaje */
	d = util_get_distance (data->x, data->y, px, py);

	d = pratio + data->ratio - d;
	
	data->y -= d * sin (angle);
	data->x += d * cos (angle);
}
